/*
 * input_method.c - taiwins server input-method implementation
 *
 * Copyright (c) 2020 Xichen Zhou
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include <assert.h>
#include <stdint.h>
#include <string.h>
#include <unistd.h>
#include <sys/mman.h>
#include <wayland-server-core.h>
#include <wayland-server.h>

#include <taiwins/objects/input_method.h>
#include <taiwins/objects/surface.h>
#include <taiwins/objects/subsurface.h>
#include <taiwins/objects/text_input.h>
#include <taiwins/objects/seat.h>
#include <taiwins/objects/utils.h>
#include <taiwins/objects/logger.h>
#include <wayland-input-method-server-protocol.h>
#include <ctypes/os/os-compatibility.h>
#include <wayland-util.h>

static const struct zwp_input_method_v2_interface im_v2_impl;
static const struct zwp_input_method_keyboard_grab_v2_interface grab_impl;
static const struct zwp_input_popup_surface_v2_interface popup_impl;

static inline struct tw_input_method *
input_method_from_resource(struct wl_resource *resource)
{
	assert(wl_resource_instance_of(resource,
	                               &zwp_input_method_v2_interface,
	                               &im_v2_impl));
	return wl_resource_get_user_data(resource);
}

static inline struct tw_input_method *
input_method_from_grab_resource(struct wl_resource *resource)
{
	assert(wl_resource_instance_of(
		       resource, &zwp_input_method_keyboard_grab_v2_interface,
		       &grab_impl));
	return wl_resource_get_user_data(resource);
}

static inline struct tw_input_method *
input_method_from_popup_resource(struct wl_resource *resource)
{
	assert(wl_resource_instance_of(resource,
	                               &zwp_input_popup_surface_v2_interface,
	                               &popup_impl));
	return wl_resource_get_user_data(resource);

}

static inline bool
input_method_has_popup(struct tw_input_method *im)
{
	//simply check if we have subsurface
	return im->im_surface.subsurface.surface != NULL;
}

/******************************************************************************
 * input method popup implemenation
 *****************************************************************************/

static void
commit_input_popup_surface(struct tw_surface *surface)
{
	struct tw_subsurface *sub = surface->role.commit_private;
	struct tw_surface *parent = sub->parent;
	struct tw_input_method *im =
		wl_container_of(sub, im, im_surface.subsurface);

        if (!parent)
		return;
        //set position same as subsurface.
        tw_surface_set_position(sub->surface,
                                sub->parent->geometry.x+sub->sx,
                                sub->parent->geometry.y+sub->sy);
}

static const struct zwp_input_popup_surface_v2_interface popup_impl = {
	.destroy = tw_resource_destroy_common,
};

static struct tw_surface_role tw_input_popup_role = {
	.commit = commit_input_popup_surface,
	.name = "input-method-popup",
	.link.prev = &tw_input_popup_role.link,
	.link.next = &tw_input_popup_role.link,
};

/**
 * calculate the relative position of the text input from input method surface
 *
 * The text_cursor here is in the parent surface space.
 */
static void
im_popup_calc_pos(struct tw_input_method *im,
                  const pixman_rectangle32_t *text_cursor)
{
	struct tw_subsurface *sub = &im->im_surface.subsurface;
	if (!sub->surface || !sub->resource || !sub->parent)
		return;

	//The simplest case is just set it underneath.
	sub->sx = text_cursor->x + text_cursor->width;
	sub->sy = text_cursor->y + text_cursor->height;

	zwp_input_popup_surface_v2_send_text_input_rectangle(
		sub->resource, -text_cursor->width,
		-text_cursor->height, text_cursor->width,
		text_cursor->height);
}

static void
input_method_disable(struct tw_input_method *im, struct wl_resource *focused)
{
	struct tw_subsurface *sub = &im->im_surface.subsurface;

	//don't disable if not focused (already disabled)
	if (im->text_input.focused != focused)
		return;
	tw_logl_level(TW_LOG_DBUG, "input method popup disable");

	im->text_input.focused = NULL;
	tw_reset_wl_list(&im->text_input.destroy.link);
	zwp_input_method_v2_send_deactivate(im->resource);
	//also unmap subsurface if possible
	if (input_method_has_popup(im))
		tw_subsurface_hide(sub);
}

static void
notify_im_text_input_destroy(struct wl_listener *listener, void *data)
{
	struct tw_input_method *im =
		wl_container_of(listener, im, text_input.destroy);
	assert(data == im->text_input.focused);
	input_method_disable(im, data);
}


/*
 * try enabling the input-method popup.
 *
 * Surface and resource is set in handle_input_method_get_input_popup_surface;
 * And getting a parent from the list of text_inputs. If all those requirements
 * are met, we are ready to show the popup surface. Otherwise the popup is not
 * shown.
 *
 * The first time we were asked to enable the popup surface, the popup is not
 * created yet, im_popup_enable is called again later on get_popup_surface
 */
static void
input_method_enable(struct tw_input_method *im, struct wl_resource *focused)
{
	struct tw_subsurface *sub = &im->im_surface.subsurface;
	struct tw_text_input *ti = focused ?
		tw_text_input_from_resource(focused) : NULL;
	struct tw_surface *parent = focused ? ti->focused : NULL;

	tw_logl_level(TW_LOG_DBUG, "input method popup enable");

	//TODO: if equal, we risk adding resource destroy listener twice
	if (!focused || !parent)
		return;
	if (im->text_input.focused != focused) {
		input_method_disable(im, im->text_input.focused);
		tw_reset_wl_list(&im->text_input.destroy.link);
		im->text_input.focused = focused;
		tw_set_resource_destroy_listener(focused,
		                                 &im->text_input.destroy,
		                                 notify_im_text_input_destroy);
	zwp_input_method_v2_send_activate(im->resource);
	}

	//adding popup surface if available
	if (input_method_has_popup(im))
		tw_subsurface_show(sub, parent);
}

static inline void
destroy_im_popup(struct tw_input_method *im)
{
	struct tw_subsurface *sub = &im->im_surface.subsurface;
	if (input_method_has_popup(im))
		tw_subsurface_fini(sub);
}

static void
destroy_im_popup_resource(struct wl_resource *resource)
{
	struct tw_input_method *im =
		input_method_from_popup_resource(resource);

	if (!im)
		return;
	//client should not destroy the surface before the popup destroy, we
	//will notify client for that
	if (!input_method_has_popup(im)) {
		wl_resource_post_error(resource, 0, "input method should not "
		                       "destroy surface before destroying "
		                       "input method popup object@%d",
		                       wl_resource_get_id(resource));
	}
	im->im_surface.subsurface.resource = NULL;
}

/* this should be called after destroy_im_popup_resource */
static void
notify_im_popup_surface_destroyed(struct wl_listener *listener, void *data)
{
	struct tw_subsurface *sub =
		wl_container_of(listener, sub, surface_destroyed);
	struct tw_input_method *im =
		wl_container_of(sub, im, im_surface.subsurface);

	destroy_im_popup(im);
}

/******************************************************************************
 * input method keyboard implemenation
 *****************************************************************************/

static const struct zwp_input_method_keyboard_grab_v2_interface grab_impl = {
	.release = tw_resource_destroy_common,
};

static void
notify_im_keyboard_key(struct tw_seat_keyboard_grab *grab, uint32_t time_msec,
                       uint32_t key, uint32_t state)
{
	struct wl_display *display = grab->seat->display;
	uint32_t serial = wl_display_next_serial(display);

	zwp_input_method_keyboard_grab_v2_send_key(grab->data, serial,
	                                           time_msec, key, state);
}

static void
notify_im_keyboard_modifiers(struct tw_seat_keyboard_grab *grab,
                             uint32_t mods_depressed, uint32_t mods_latched,
                             uint32_t mods_locked, uint32_t group)
{
	struct wl_display *display = grab->seat->display;
	uint32_t serial = wl_display_next_serial(display);

	zwp_input_method_keyboard_grab_v2_send_modifiers(grab->data, serial,
	                                                 mods_depressed,
	                                                 mods_latched,
	                                                 mods_locked, group);
}

static const struct tw_keyboard_grab_interface im_grab_impl = {
	.key = notify_im_keyboard_key,
	.modifiers = notify_im_keyboard_modifiers,
	.enter = tw_keyboard_default_enter,
	.cancel = tw_keyboard_default_cancel,
};

static void
input_method_send_keymap(struct tw_input_method *im,
                         struct tw_keyboard *keyboard,
                         struct wl_resource *grab_resource)
{
	int keymap_fd;
	void *ptr;

	//this would later be a problem for virtual keyboards to involving
	//looping keymap sending.
	if (keyboard->keymap_string) {
		keymap_fd = os_create_anonymous_file(keyboard->keymap_size);
		if (!keymap_fd)
			return;
		ptr = mmap(NULL, keyboard->keymap_size, PROT_READ | PROT_WRITE,
		           MAP_SHARED, keymap_fd, 0);
		if (ptr == MAP_FAILED)
			goto close_fd;
		strcpy(ptr, keyboard->keymap_string);
		zwp_input_method_keyboard_grab_v2_send_keymap(
			grab_resource, WL_KEYBOARD_KEYMAP_FORMAT_XKB_V1,
			keymap_fd, keyboard->keymap_size);
		munmap(ptr, keyboard->keymap_size);
	close_fd:
		close(keymap_fd);
	}
}

static void
input_method_start_grab(struct tw_input_method *im,
                        struct tw_keyboard *keyboard,
                        struct wl_resource *grab_resource)
{
	struct tw_seat *seat = wl_container_of(keyboard, seat, keyboard);

	im->im_grab.data = grab_resource;
	im->im_grab.impl = &im_grab_impl;

	input_method_send_keymap(im, keyboard, grab_resource);
	zwp_input_method_keyboard_grab_v2_send_repeat_info(grab_resource,
	                                                   seat->repeat_rate,
	                                                   seat->repeat_delay);
	tw_keyboard_start_grab(keyboard, &im->im_grab,
	                       TW_INPUT_METHOD_GRAB_ORDER);
}

static void
destroy_im_grab(struct tw_input_method *im)
{
	struct tw_seat *seat = im->seat;
	struct wl_resource *resource = im->im_grab.data;

	tw_keyboard_end_grab(&seat->keyboard, &im->im_grab);
	if (resource)
		wl_resource_set_user_data(resource, NULL);
	im->im_grab.data = NULL;
}

static void
handle_destroy_im_grab_resource(struct wl_resource *resource)
{
	struct tw_input_method *im = input_method_from_grab_resource(resource);
	if (im)
		destroy_im_grab(im);
}

/******************************************************************************
 * input method implemenation
 *****************************************************************************/

static void
handle_input_method_commit_string(struct wl_client *client,
                                  struct wl_resource *resource,
                                  const char *text)
{
	struct tw_input_method *im = input_method_from_resource(resource);
	if (!im)
		return;
	free(im->pending.commit_string);
	im->pending.commit_string = strdup(text);

	if (!im->pending.commit_string)
		wl_resource_post_no_memory(resource);

	im->pending.requests |= TW_INPUT_METHOD_COMMIT_STRING;
}

static void
handle_input_method_set_preedit_string(struct wl_client *client,
				   struct wl_resource *resource,
				   const char *text,
				   int32_t cursor_begin,
				   int32_t cursor_end)
{
	struct tw_input_method *im = input_method_from_resource(resource);
	if (!im)
		return;
	free(im->pending.preedit.text);
	im->pending.preedit.text = strdup(text);

	if (!im->pending.preedit.text)
		wl_resource_post_no_memory(resource);

	im->pending.preedit.cursor_begin = cursor_begin;
	im->pending.preedit.cursor_end = cursor_end;
	im->pending.requests |= TW_INPUT_METHOD_PREEDIT;
}

static void
handle_input_method_delete_surrounding_text(struct wl_client *client,
                                             struct wl_resource *resource,
                                             uint32_t before_length,
                                             uint32_t after_length)
{
	struct tw_input_method *im = input_method_from_resource(resource);
	if (!im)
		return;
	im->pending.surrounding_delete.before_length = before_length;
	im->pending.surrounding_delete.after_length = after_length;
	im->pending.requests |= TW_INPUT_METHOD_SURROUNDING_DELETE;
}

static void
handle_input_method_commit(struct wl_client *client,
                           struct wl_resource *resource,
                           uint32_t serial)
{
	struct tw_input_method *im = input_method_from_resource(resource);
	struct tw_input_method_state reset = {0};
	struct tw_text_input *ti = (im && im->text_input.focused) ?
		tw_text_input_from_resource(im->text_input.focused) : NULL;

	if (ti) {
		struct tw_text_input_event e = {0};

		if (im->pending.requests & TW_INPUT_METHOD_PREEDIT) {
			e.preedit.text = im->pending.preedit.text;
			e.preedit.cursor_begin =
				im->pending.preedit.cursor_begin;
			e.preedit.cursor_end =
				im->pending.preedit.cursor_end;
			e.events |= TW_TEXT_INPUT_PREEDIT;
		}
		if (im->pending.requests & TW_INPUT_METHOD_COMMIT_STRING) {
			e.commit_string = im->pending.commit_string;
			e.events |= TW_TEXT_INPUT_COMMIT_STRING;
		}
		if (im->pending.requests & TW_INPUT_METHOD_SURROUNDING_DELETE){
			e.surrounding_delete.before_length =
				im->pending.surrounding_delete.before_length;
			e.surrounding_delete.after_length =
				im->pending.surrounding_delete.after_length;
			e.events |= TW_TEXT_INPUT_SURROUNDING_DELETE;
		}
		tw_text_input_commit_event(ti, &e);
	}

	//swap state
	free(im->current.commit_string);
	free(im->current.preedit.text);
	im->current = im->pending;
	im->pending = reset;
}


static void
handle_input_method_get_input_popup_surface(struct wl_client *client,
                                            struct wl_resource *resource,
                                            uint32_t id,
                                            struct wl_resource *surf_resource)
{
	struct wl_resource *popup_resource;
	struct tw_input_method *im = input_method_from_resource(resource);
	uint32_t version = wl_resource_get_version(resource);
	struct tw_surface *surface = tw_surface_from_resource(surf_resource);

	if (!tw_surface_assign_role(surface, &tw_input_popup_role,
	                            &im->im_surface.subsurface)) {
		wl_resource_post_error(resource, -1, "failed to set input "
		                       "method popup, wl_surface@%d already "
		                       "has a role",
		                       wl_resource_get_id(surface->resource));
	}
	if (!im)
		return;
	tw_logl_level(TW_LOG_DBUG, "input method get popup");
	popup_resource =
		wl_resource_create(client,
		                   &zwp_input_popup_surface_v2_interface,
		                   version, id);
	if (!popup_resource) {
		wl_resource_post_no_memory(resource);
		return;
	}
	wl_resource_set_implementation(popup_resource, &popup_impl, im,
	                               destroy_im_popup_resource);
	tw_subsurface_init(&im->im_surface.subsurface, popup_resource, surface,
	                   NULL, notify_im_popup_surface_destroyed);
	input_method_enable(im, im->text_input.focused);
}

static void
handle_input_method_grab_keyboard(struct wl_client *client,
                                  struct wl_resource *resource,
                                  uint32_t keyboard)
{
	struct tw_seat *seat;
	struct wl_resource *grab_resource = NULL;
	uint32_t version = wl_resource_get_version(resource);
	struct tw_input_method *im = input_method_from_resource(resource);

	if (!im)
		return;
	seat = im->seat;
	//no keyboard
	if (!(seat->capabilities & WL_SEAT_CAPABILITY_KEYBOARD))
		return;
	//already in grab, delete the old one
	if (seat->keyboard.grab == &im->im_grab ||
	    seat->keyboard.grab != &seat->keyboard.default_grab)
		destroy_im_grab(im);
	if (!(grab_resource =
	      wl_resource_create(client,
	                         &zwp_input_method_keyboard_grab_v2_interface,
	                         version, keyboard))) {
		wl_resource_post_no_memory(resource);
		return;
	}
	//a keyboard grab.
	wl_resource_set_implementation(grab_resource, &grab_impl, im,
	                               handle_destroy_im_grab_resource);
	input_method_start_grab(im, &seat->keyboard, grab_resource);
}

static const struct zwp_input_method_v2_interface im_v2_impl = {
	.commit_string = handle_input_method_commit_string,
	.set_preedit_string = handle_input_method_set_preedit_string,
	.delete_surrounding_text = handle_input_method_delete_surrounding_text,
	.commit = handle_input_method_commit,
	.get_input_popup_surface = handle_input_method_get_input_popup_surface,
	.grab_keyboard = handle_input_method_grab_keyboard,
	.destroy = tw_resource_destroy_common,
};

static void
destroy_input_method(struct tw_input_method *im)
{
	wl_list_remove(&im->link);
	wl_list_remove(&im->seat_destroy_listener.link);
	wl_list_remove(wl_resource_get_link(im->resource));
	wl_resource_set_user_data(im->resource, NULL);
	input_method_disable(im, im->text_input.focused);
	destroy_im_popup(im);
	destroy_im_grab(im);

	free(im->current.commit_string);
	free(im->current.preedit.text);

	free(im);
}

static void
destroy_input_method_resource(struct wl_resource *resource)
{
	struct tw_input_method *im = input_method_from_resource(resource);
	if (!im)
		return;
	destroy_input_method(im);
}

static void
notify_im_v2_seat_destroy(struct wl_listener *listener, void *data)
{
	struct tw_input_method *im =
		wl_container_of(listener, im, seat_destroy_listener);
	zwp_input_method_v2_send_unavailable(im->resource);
	destroy_input_method(im);
}

/******************************************************************************
 * input method manager
 *****************************************************************************/

static const struct zwp_input_method_manager_v2_interface im_manager_impl;

static inline struct tw_input_method_manager *
manager_from_resource(struct wl_resource *resource)
{
	assert(wl_resource_instance_of(resource,
	                               &zwp_input_method_manager_v2_interface,
	                               &im_manager_impl));
	return wl_resource_get_user_data(resource);
}

static void
im_manager_handle_get_input_method(struct wl_client *client,
                                   struct wl_resource *manager_resource,
                                   struct wl_resource *seat_resource,
                                   uint32_t input_method)
{
	struct tw_input_method *im;
	struct wl_resource *im_resource = NULL;
	uint32_t version = wl_resource_get_version(manager_resource);
	struct tw_seat *seat = tw_seat_from_resource(seat_resource);
	struct tw_input_method_manager *manager =
		manager_from_resource(manager_resource);


	if (!tw_create_wl_resource_for_obj(im_resource, im, client,
	                                   input_method, version,
	                                   zwp_input_method_v2_interface)) {
		wl_resource_post_no_memory(manager_resource);
		return;
	}
	im->seat = seat;
	im->resource = im_resource;
	im->text_input.focused = NULL;
	wl_list_init(&im->link);
	wl_list_init(&im->text_input.destroy.link);
	wl_resource_set_implementation(im_resource, &im_v2_impl, im,
	                               destroy_input_method_resource);
	tw_signal_setup_listener(&seat->signals.destroy,
	                         &im->seat_destroy_listener,
	                         notify_im_v2_seat_destroy);
	//as described in protocol, if we have an input method already in the
	//seat, the only thing we need to do is sending unavailable event
	if (tw_input_method_find_from_seat(seat)) {
		zwp_input_method_v2_send_unavailable(im->resource);
		destroy_input_method(im);
		return;
	}
	wl_list_insert(manager->ims.prev, &im->link);
	wl_list_insert(seat->resources.prev,
	               wl_resource_get_link(im_resource));
}

static const struct zwp_input_method_manager_v2_interface im_manager_impl = {
	.get_input_method = im_manager_handle_get_input_method,
	.destroy = tw_resource_destroy_common,
};

static void
bind_input_method_manager(struct wl_client *wl_client, void *data,
                          uint32_t version, uint32_t id)
{
	struct wl_resource *resource =
		wl_resource_create(wl_client,
		                   &zwp_input_method_manager_v2_interface,
		                   version, id);
	if (!resource) {
		wl_client_post_no_memory(wl_client);
		return;
	}
	wl_resource_set_implementation(resource, &im_manager_impl, data,
	                               NULL);
}

static void
notify_im_manager_display_destroy(struct wl_listener *listener, void *data)
{
	struct tw_input_method_manager *manager =
		wl_container_of(listener, manager, display_destroy_listener);

	assert(data == manager->display);
	wl_global_destroy(manager->global);
	manager->global = NULL;
	manager->display = NULL;
	tw_reset_wl_list(&listener->link);
}

WL_EXPORT void
tw_input_method_send_event(struct tw_input_method *im,
                           struct tw_input_method_event *e)
{
	if ((e->events & TW_INPUT_METHOD_TOGGLE) && e->enabled)
		input_method_enable(im, e->focused);
	else if (e->events & TW_INPUT_METHOD_TOGGLE)
		input_method_disable(im, e->focused);

	if (e->events & TW_INPUT_METHOD_SURROUNDING_TEXT)
		zwp_input_method_v2_send_surrounding_text(
			im->resource, e->surrounding.text,
			e->surrounding.cursor, e->surrounding.anchor);
	if (e->events & TW_INPUT_METHOD_CHANGE_CAUSE)
		zwp_input_method_v2_send_text_change_cause(im->resource,
		                                           e->change_cause);
	if (e->events & TW_INPUT_METHOD_CONTENT_TYPE)
		zwp_input_method_v2_send_content_type(im->resource,
		                                      e->content_hint,
		                                      e->content_purpose);
	if (e->events & TW_INPUT_METHOD_CURSOR_RECTANGLE) {
		//sending rectanges if it there
		im_popup_calc_pos(im, &e->cursor_rect);
		im->im_surface.rectangle = e->cursor_rect;
	}
	zwp_input_method_v2_send_done(im->resource);
}

WL_EXPORT struct tw_input_method *
tw_input_method_find_from_seat(struct tw_seat *seat)
{
	struct tw_input_method *method;
	struct wl_resource *res;

	wl_resource_for_each(res, &seat->resources) {
		if (wl_resource_instance_of(res,
		                            &zwp_input_method_v2_interface,
		                            &im_v2_impl) &&
		    (method = input_method_from_resource(res)))
			return method;
	}
	return NULL;
}

WL_EXPORT bool
tw_input_method_manager_init(struct tw_input_method_manager *manager,
                             struct wl_display *display)
{
	if (!(manager->global =
	      wl_global_create(display, &zwp_input_method_manager_v2_interface,
	                       1, manager, bind_input_method_manager)))
		return false;

	manager->display = display;
	tw_set_display_destroy_listener(display,
	                                &manager->display_destroy_listener,
	                                notify_im_manager_display_destroy);
	wl_list_init(&manager->ims);
	tw_subsurface_add_role(&tw_input_popup_role);
	return true;
}

WL_EXPORT struct tw_input_method_manager *
tw_input_method_manager_create_global(struct wl_display *display)
{
	static struct tw_input_method_manager s_manager = {0};
	assert(s_manager.display == NULL || s_manager.display == display);
	if (s_manager.display == display)
		return &s_manager;
	else if (tw_input_method_manager_init(&s_manager, display))
		return &s_manager;
	return NULL;
}
